/*
 *  Copyright 2006 Adrian Thurston <thurston@cs.queensu.ca>
 */

/*  This file is part of Kelbt.
 *
 *  Kelbt is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Kelbt is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Kelbt; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#include <iostream>
#include <fstream>
#include <string.h>

#include "kelbt.h"
#include "klparse.h"

using std::ifstream;
using std::istream;
using std::ostream;
using std::cout;
using std::cerr;
using std::endl;

void do_scan( char *fileName, istream &input, ParserDict &parserDict, int include_depth );

%%{
	machine klscan;

	# This is sent by the driver code.
	EOF = 0;

	# Identifiers, numbers, commetns, and other common things.
	ident = ( alpha | '_' ) ( alpha |digit |'_' )*;
	number = digit+;
	c_comment = '/*' any* :>> '*/' | '//' [^\n]* '\n';
	literal = '"' ([^"\\] | '\\' any)* '"' | "'" ([^'\\] | '\\' any)* "'";
	whitespace = [ \t\n];
	pound_comment = '#' [^\n]* '\n';

	# An inline block of code. This is specified as a scanned, but is sent to
	# the parser as one long block. The inline_block pointer is used to handle
	# the preservation of the data.
	inline_code := |*
		ident; 
		number;
		c_comment;
		literal;
		whitespace+;

		'{' => { curly_count += 1; };

		'}' => { 
			if ( --curly_count == 0 ) {
				/* Send the entire block and free the inline_start pointer. */
				parser->token( TK_Inline, inline_start, tokend );
				inline_start = 0;
				fgoto parser_def;
			}
		};

		( '$' | '@' ) number | '$$' | '@@' => {
			if ( inline_start < tokstart )
				parser->token( TK_Inline, inline_start, tokstart );
			parser->token( TK_Reference, tokstart, tokend );
			inline_start = tokend;
		};

		any;
	*|;

	action handle_parser
	{
		//cout << "PARSER STATEMENT" << endl;

		char *data = identstart;
		int length = identend - identstart;
		char *parserName = new char[length+1];
		memcpy( parserName, data, length );
		parserName[length] = 0;

		pdEl = parserDict.find( parserName );
		if ( pdEl != 0 ) {
			//cout << "USING EXISTING PARSER" << endl;
			delete[] parserName;
		}
		else {
			//cout << "CREATING NEW PARSER" << endl;
			pdEl = new ParserDictEl;
			pdEl->name = parserName;
			pdEl->parser = new Parser( fileName, parserName );
			pdEl->parser->init();
			parserDict.insert( pdEl );
		}

		parser = pdEl->parser;
		parser->startSection();
	}

	action handle_include
	{
		char *data = litstart + 1;
		int length = litend - litstart - 1;
		char *includeFileName = new char[length+1];
		memcpy( includeFileName, data, length );
		includeFileName[length] = 0;

		//cout << "HIT INCLUDE " << includeFileName << endl;

		/* Open the input file for reading. */
		ifstream *inFile = new ifstream( includeFileName );
		istream *inStream = inFile;
		if ( ! inFile->is_open() )
			error() << "include: could not open " << includeFileName << " for reading" << endl;

		do_scan( includeFileName, *inStream, parserDict, include_depth+1 );

		delete[] includeFileName;
		delete inFile;
	}

	# Parser definitions. 
	parser_def := |*
		'commit' => { parser->token( KW_Commit, 0, 0 ); };
		'try' => { parser->token( KW_Try, 0, 0 ); };
		'undo' => { parser->token( KW_Undo, 0, 0 ); };
		'final' => { parser->token( KW_Final, 0, 0 ); };
		'translate' => { parser->token( KW_Translate, 0, 0 ); };
		'token' => { parser->token( KW_Token, 0, 0 ); };
		'nonterm' => { parser->token( KW_NonTerm, 0, 0 ); };
		'uses' => { parser->token( KW_Uses, 0, 0 ); };
		'type' => { parser->token( KW_Type, 0, 0 ); };
		'interface' => { parser->token( KW_Interface, 0, 0 ); };

		# In the parser statement mark the start and end of the parser name.
		# Since we are within a token, we know these will be valid until the
		# include call.
		parser_ident = ident
			>{ identstart = fpc; }
			%{ identend = fpc; };

		'parser' [ \t]* parser_ident [ \t]* ';'
				=> handle_parser;

		# Mark the start and end of the include literal.
		include_lit = literal 
			>{ litstart = fpc; }
			@{ litend = fpc; };

		'include' [ \t]* include_lit [ \t]* ';'
				=> handle_include;

		# Parser literals accept are single characters with single quotes.
		parser_literal = "'" ([^'\\] | '\\' any) "'";

		ident => { parser->token( TK_Word, tokstart, tokend ); } ;
		parser_literal => { parser->token( TK_Literal, tokstart+1, tokend-1 ); } ;
		pound_comment;

		'}%%' => { 
			/* In order to generate anything we must be in the top level file
			 * and the current spec must be active and there must not have been
			 * any parse errors. */
			if ( include_depth == 0 ) 
				parser->endSection();
			fgoto main;
		};

		whitespace+;

		'{' => { 
			inline_start = tokstart;
			curly_count = 1; 
			fgoto inline_code;
		};

		any => { parser->token( *tokstart, 0, 0 ); } ;
	*|;

	action pass {
		if ( !generateGraphviz && gblErrorCount == 0 && include_depth == 0 )
			outStream->write( tokstart, tokend-tokstart );
	}

	# Outside code scanner. These tokens get passed through.
	main := |*
		ident => pass;
		number => pass;
		c_comment => pass;
		literal => pass;
		'%%{' => { fgoto parser_def; };
		whitespace+ => pass;
		EOF;
		any => pass;
	*|;

}%%

%% write data;

void do_scan( char *fileName, istream &input, ParserDict &parserDict, int include_depth )
{
	int bufsize = 8;
	char *buf = new char[bufsize];
	const char last_char = 0;
	char *tokstart, *tokend;
	int cs, act, have = 0;
	int curly_count = 0;
	char *inline_start = 0;
	char *litstart = 0, *litend = 0;
	char *identstart = 0, *identend = 0;
	ParserDictEl *pdEl = 0;
	Parser *parser = 0;
	bool execute = true;

	%% write init;

	while ( execute ) {
		char *p = buf + have;
		int space = bufsize - have;

		if ( space == 0 ) {
			/* We filled up the buffer trying to scan a token. Grow it. */
			bufsize = bufsize * 2;
			char *newbuf = new char[bufsize];
			//cout << "FULL BUFFER, NEW SIZE: " << bufsize << endl;

			/* Recompute p and space. */
			p = newbuf + have;
			space = bufsize - have;

			/* Patch up pointers possibly in use. */
			if ( tokstart != 0 )
				tokstart = newbuf + ( tokstart - buf );
			if ( inline_start != 0 )
				inline_start = newbuf + ( inline_start - buf );
			tokend = newbuf + ( tokend - buf );
			litstart = newbuf + ( litstart - buf );
			litend = newbuf + ( litend - buf );
			identstart = newbuf + ( identstart - buf );
			identend = newbuf + ( identend - buf );

			/* Copy the new buffer in. */
			memcpy( newbuf, buf, have );
			delete[] buf;
			buf = newbuf;
		}

		input.read( p, space );
		int len = input.gcount();

		/* If we see eof then append the EOF char. */
	 	if ( len == 0 ) {
			p[0] = last_char, len = 1;
			execute = false;
		}

		char *pe = p + len;
		%% write exec;

		/* Check if we failed. */
		if ( cs == klscan_error ) {
			/* Machine failed before finding a token. */
			cout << "PARSE ERROR" << endl;
			exit(1);
		}

		/* Decide if we need to preserve anything. */
		char *preserve = tokstart;
		if ( inline_start != 0 && ( preserve == 0 || inline_start < preserve ) )
			preserve = inline_start;

		/* Now set up the prefix. */
		if ( preserve == 0 )
			have = 0;
		else {
			/* There is data that needs to be shifted over. */
			have = pe - preserve;
			memmove( buf, preserve, have );
			unsigned int shiftback = preserve - buf;
			if ( tokstart != 0 )
				tokstart -= shiftback;
			if ( inline_start != 0 )
				inline_start -= shiftback;

			tokend -= shiftback;
			litstart -= shiftback;
			litend -= shiftback;
			identstart -= shiftback;
			identend -= shiftback;

			preserve = buf;
		}
	}

	delete[] buf;
}

void scan( char *fileName, istream &input )
{
	ParserDict parserDict;
	do_scan( fileName, input, parserDict, 0 );
}
